from typing import Union
from dataclasses import asdict, dataclass
from transformers import (AutoTokenizer, AutoModelForCausalLM, GenerationConfig)
import torch


from .base_textgen import TextGenerator
from ...datamodel import TextGenerationConfig, TextGenerationResponse
from ...utils import cache_request


@dataclass
class DialogueTemplate:
    system: str = None
    dialogue_type: str = "default"
    messages: list[dict[str, str]] = None
    system_token: str = "<|system|>"
    user_token: str = "<|user|>"
    assistant_token: str = "<|assistant|>"
    end_token: str = "<|end|>"

    def get_inference_prompt(self) -> str:
        if self.dialogue_type == "default":
            prompt = ""
            system_prompt = (
                self.system_token + "\n" + self.system + self.end_token + "\n"
                if self.system
                else ""
            )
            if self.messages is None:
                raise ValueError("Dialogue template must have at least one message.")
            for message in self.messages:
                if message["role"] == "system":
                    system_prompt += (
                        self.system_token
                        + "\n"
                        + message["content"]
                        + self.end_token
                        + "\n"
                    )
                elif message["role"] == "user":
                    prompt += (
                        self.user_token
                        + "\n"
                        + message["content"]
                        + self.end_token
                        + "\n"
                    )
                else:
                    prompt += (
                        self.assistant_token
                        + "\n"
                        + message["content"]
                        + self.end_token
                        + "\n"
                    )
            prompt += self.assistant_token
            if system_prompt:
                prompt = system_prompt + prompt
            return prompt
        elif self.dialogue_type == "alpaca":
            prompt = (
                self.user_token + "\n" + (self.system + "\n" if self.system else "")
            )
            for message in self.messages:
                prompt += message["content"] + "\n"
            prompt = prompt + " " + self.assistant_token + "\n"
            # print(instruction)
            return prompt
        elif self.dialogue_type == "llama2":
            prompt = "[INST]"
            system_prompt = ""
            other_prompt = ""

            for message in self.messages:
                if message["role"] == "system":
                    system_prompt += message["content"] + "\n"
                elif message["role"] == "assistant":
                    other_prompt += message["content"] + "  \n"
                else:
                    other_prompt += "[INST] " + message["content"] + "[/INST]\n"

            prompt = (
                prompt
                + f" <<SYS>> {system_prompt} <</SYS>> \n"
                + other_prompt
                + "[/INST]"
            )


class HFTextGenerator(TextGenerator):
    def __init__(self, provider: str = "huggingface", device_map: str = "auto", **kwargs):

        super().__init__(provider=provider)

        self.dialogue_type = kwargs.get("dialogue_type", "alpaca")

        self.model_name = kwargs.get("model", "TheBloke/gpt4-x-vicuna-13B-HF")
        self.load_in_8bit = kwargs.get("load_in_8bit", False)
        self.device = kwargs.get("device", self.get_default_device())
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name, trust_remote_code=True)
        self.model = AutoModelForCausalLM.from_pretrained(
            self.model_name, 
            device_map=device_map, 
            load_in_8bit=self.load_in_8bit, 
            trust_remote_code=True
        )
        if not device_map:
            self.model.to(self.device)
        self.model.config.pad_token_id = self.tokenizer.pad_token_id

        self.max_length = kwargs.get("max_length", 1024*8)
        self.max_context_tokens = kwargs.get(
            "max_context_tokens", self.model.config.max_position_embeddings
        )

        if self.dialogue_type == "alpaca":
            self.dialogue_template = DialogueTemplate(
                dialogue_type="alpaca",
                end_token=self.tokenizer.eos_token,
                user_token="### Instruction:",
                assistant_token="### Response:",
            )
            self.model.config.pad_token_id = self.tokenizer.pad_token_id = 0  # unk
            self.model.config.bos_token_id = 1
            self.model.config.eos_token_id = 2
        else:
            self.dialogue_template = DialogueTemplate(end_token=self.tokenizer.eos_token)

    def get_default_device(self):
        """Pick GPU if available, else CPU"""
        if torch.cuda.is_available():
            return torch.device("cuda")
        elif torch.backends.mps.is_available():
            return torch.device("mps")
        else:
            return torch.device("cpu")

    def post_process_response(self, response):
        response = (
            response.split(self.dialogue_template.assistant_token)[-1]
            .replace(self.dialogue_template.end_token, "")
            .strip()
        )
        response = {"role": "assistant", "content": response}
        return response

    def messages_to_instruction(self, messages):
        instruction = "### Instruction: "
        for message in messages:
            instruction += message["content"] + "\n"
        instruction = instruction + "### Response: "
        # print(instruction)
        return instruction

    def generate(
            self, messages: Union[list[dict],
                                  str],
            config: TextGenerationConfig = TextGenerationConfig(),
            **kwargs) -> TextGenerationResponse:
        use_cache = config.use_cache
        config.model = self.model_name
        cache_key_params = {
            **asdict(config),
            **kwargs,
            "messages": messages,
            "dialogue_type": self.dialogue_type}
        if use_cache:
            response = cache_request(cache=self.cache, params=(cache_key_params))
            if response:
                return TextGenerationResponse(**response)

        self.dialogue_template.messages = messages
        prompt = self.dialogue_template.get_inference_prompt()
        # print("prompt:", prompt)
        # print('\n')
        batch = self.tokenizer(
            prompt, return_tensors="pt", return_token_type_ids=False
        ).to(self.model.device)
        input_ids = batch["input_ids"]

        max_new_tokens = kwargs.get(
            "max_new_tokens", self.max_context_tokens - input_ids.shape[-1]
        )
        # print(
        #     "***********Prompt tokens: ",
        #     input_ids.shape[-1],
        #     " | Max new tokens: ",
        #     max_new_tokens,
        # )

        top_k = kwargs.get("top_k", config.top_k)
        min_new_tokens = kwargs.get("min_new_tokens", 32)
        repetition_penalty = kwargs.get("repetition_penalty", 1.0)

        gen_config = GenerationConfig(
            max_new_tokens=max_new_tokens,
            temperature=max(config.temperature, 0.01),
            top_p=config.top_p,
            top_k=top_k,
            num_return_sequences=config.n,
            do_sample=True,
            pad_token_id=self.tokenizer.eos_token_id,
            eos_token_id=self.tokenizer.eos_token_id,
            min_new_tokens=min_new_tokens,
            repetition_penalty=repetition_penalty,
        )
        with torch.no_grad():
            self.model.eval()
            generated_ids = self.model.generate(**batch, generation_config=gen_config)

        text_response = self.tokenizer.batch_decode(
            generated_ids, skip_special_tokens=False
        )

        # print(text_response, "*************")
        prompt_tokens = len(batch["input_ids"][0])
        total_tokens = 0
        for row in generated_ids:
            total_tokens += len(row)

        usage = {
            "prompt_tokens": prompt_tokens,
            "completion_tokens": total_tokens - prompt_tokens,
            "total_tokens": total_tokens,
        }

        response = TextGenerationResponse(
            text=[self.post_process_response(x) for x in text_response],
            logprobs=[],
            config=config,
            usage=usage,
        )
        # if use_cache:
        cache_request(cache=self.cache, params=(cache_key_params), values=asdict(response))
        return response

    def count_tokens(self, text: str):
        return len(self.tokenizer(text)["input_ids"])



import os
from typing import Union, List, Dict
from diskcache import Cache
from ...utils import get_user_cache_dir
from ...datamodel import TextGenerationConfig, TextGenerationResponse
from ...version import APP_NAME
from abc import ABC, abstractmethod


class TextGenerator(ABC):

    def __init__(self, provider: str = "openai", **kwargs):
        self.provider = provider
        self.model_name = kwargs.get("model_name", "gpt-3.5-turbo")

        app_name = APP_NAME
        cache_dir_default = get_user_cache_dir(app_name)
        cache_dir_based_on_model = os.path.join(cache_dir_default, self.provider, self.model_name)
        self.cache_dir = kwargs.get("cache_dir", cache_dir_based_on_model)
        self.cache = Cache(self.cache_dir)

    @abstractmethod
    def generate(
            self, messages: Union[List[Dict],
                                  str],
            config: TextGenerationConfig = TextGenerationConfig(),
            **kwargs) -> TextGenerationResponse:
        pass

    @abstractmethod
    def count_tokens(self, text) -> int:
        pass



from dataclasses import asdict
import os
from typing import Union
from .base_textgen import TextGenerator
from ...datamodel import TextGenerationConfig, TextGenerationResponse, Message
from ...utils import cache_request, gcp_request, num_tokens_from_messages, get_gcp_credentials
from ..text.providers import providers


class PalmTextGenerator(TextGenerator):
    def __init__(
        self,
        palm_key_file: str = os.environ.get("PALM_SERVICE_ACCOUNT_KEY_FILE", None),
        project_id: str = os.environ.get("PALM_PROJECT_ID", None),
        project_location=os.environ.get("PALM_PROJECT_LOCATION", "us-central1"),
        provider: str = "google",
    ):
        super().__init__(provider=provider)

        self.project_id = project_id
        self.project_location = project_location
        self.credentials = get_gcp_credentials(palm_key_file)
        self.model_list = providers[provider]["models"] if provider in providers else {}

    def format_messages(self, messages):
        palm_messages = []
        system_messages = ""
        for message in messages:
            if message["role"] == "system":
                system_messages += message["content"] + "\n"
            else:
                if not palm_messages or palm_messages[-1]['author'] != message['role']:
                    palm_message = {
                        "author": message["role"],
                        "content": message["content"],
                    }
                    palm_messages.append(palm_message)
                else:
                    palm_messages[-1]['content'] += '\n' + message['content']

        if palm_messages and len(palm_messages) % 2 == 0:
            merged_content = palm_messages[-2]['content'] + '\n' + palm_messages[-1]['content']
            palm_messages[-2]['content'] = merged_content
            palm_messages.pop()

        return system_messages, palm_messages

    def generate(
            self, messages: Union[list[dict],
                                  str],
            config: TextGenerationConfig = TextGenerationConfig(),
            **kwargs) -> TextGenerationResponse:

        use_cache = config.use_cache
        model = config.model or "codechat-bison"
        system_messages, messages = self.format_messages(messages)
        self.model_name = model

        # print("*********", messages)

        api_url = f"https://us-central1-aiplatform.googleapis.com/v1/projects/{self.project_id}/locations/{self.project_location}/publishers/google/models/{model}:predict"

        max_tokens = self.model_list[config.model] if model in self.model_list else 1024
        palm_config = {
            'temperature': config.temperature,
            'maxOutputTokens': config.max_tokens or max_tokens
        }
        palm_payload = {
            'instances': [
                {'messages': messages,
                 'context': system_messages,
                 'examples': [],
                 }
            ],
            'parameters': palm_config
        }
        # print("*********", use_cache, palm_payload)

        cache_key_params = palm_payload
        if use_cache:
            response = cache_request(cache=self.cache, params=cache_key_params)
            if response:
                return TextGenerationResponse(**response)

        palm_response = gcp_request(
            url=api_url,
            body=palm_payload,
            method="POST",
            credentials=self.credentials)

        response_text = [
            Message(
                role="assistant" if x["author"] == "1" else x["author"],
                content=x["content"],
            )
            for x in palm_response["predictions"][0]["candidates"]
        ]

        response = TextGenerationResponse(
            text=response_text,
            logprobs=[],
            config=palm_config,
            usage={
                "total_tokens": num_tokens_from_messages(
                    response_text, model=self.model_name
                )
            },
        )

        cache_request(cache=self.cache, params=(cache_key_params), values=asdict(response))
        return response

    def count_tokens(self, text) -> int:
        return num_tokens_from_messages(text)




# This file contains the list of providers and models that are available supported by LLMX.

providers = {
    "openai": {
        "name": "OpenAI",
        "description": "OpenAI's GPT-3 and GPT-4 models.",
        "models": [
            {"name": "gpt-4", "max_tokens": 8192},
            {"name": "gpt-4-0314", "max_tokens": 8192},
            {"name": "gpt-4-0613", "max_tokens": 8192},
            {"name": "gpt-4-32k", "max_tokens": 32768},
            {"name": "gpt-4-32k-0613", "max_tokens": 32768},
            {"name": "gpt-3.5-turbo", "max_tokens": 4096},
            {"name": "gpt-3.5-turbo-0301", "max_tokens": 4096},
            {"name": "gpt-3.5-turbo-16k", "max_tokens": 16384},
            {"name": "gpt-3.5-turbo-0613", "max_tokens": 4096},
            {"name": "gpt-3.5-turbo-16k-0613", "max_tokens": 16384},
        ]
    },
    "azure": {
        "name": "Azure",
        "description": "Azure's GPT-3 and GPT-4 models.",
        "models": [
            {"name": "gpt-35-turbo", "max_tokens": 8192},
            {"name": "gpt-35-turbo-16k", "max_tokens": 16384},
        ]
    },
    "google": {
        "name": "Google",
        "description": "Google's LLM models.",
        "models": [
            {"name": "codechat-bison", "max_tokens": 1024},
            {"name": "chat-bison", "max_tokens": 1024},
        ]
    },
    "cohere": {
        "name": "Cohere",
        "description": "Cohere's LLM models.",
        "models": [
            {"name": "command", "max_tokens": 4096},
            {"name": "command-nightly", "max_tokens": 4096},
        ]
    },
    "huggingface": {
        "name": "HuggingFace",
        "description": "HuggingFace's LLM models.",
        "models": [
            {"name": "TheBloke/Llama-2-7b-chat-fp16", "max_tokens": 4096},
            {"name": "TheBloke/Llama-2-13B-fp16", "max_tokens": 4096},
            {"name": "TheBloke/gpt4-x-vicuna-13B-HF", "max_tokens": 2040},
        ],
    }
}

# providers["palm"] = providers["google"]



from setuptools import setup, find_packages

setup(
    name='llmx_zk',
    version='0.1.0',
    description='Added azure GPT to the original foundation',
    packages=find_packages(),
    keywords=['python', 'linux', 'llm', 'windows'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: Microsoft :: Windows',
    ],
    install_requires=[
        # 列出项目依赖的其他 Python 包
        'requests',
        'openai',
		'google_auth', 
    ]
)
